from OpenSSL import crypto
import arrow
import md5
import rsa
import os
import base64

SIGN_ENV_TMPL = """<ds:SignedInfo xmlns:ds="http://www.w3.org/2000/09/xmldsig#"><ds:CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"></ds:CanonicalizationMethod><ds:SignatureMethod Algorithm="http://www.w3.org/2000/09/xmldsig#rsa-sha1"></ds:SignatureMethod><ds:Reference URI="#%(body_id)s"><ds:Transforms><ds:Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"></ds:Transform></ds:Transforms><ds:DigestMethod Algorithm="http://www.w3.org/2000/09/xmldsig#sha1"></ds:DigestMethod><ds:DigestValue>%(digest_value)s</ds:DigestValue></ds:Reference></ds:SignedInfo>"""

XML_TMPL = """<SOAP-ENV:Envelope xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/"><SOAP-ENV:Header><wsse:Security xmlns:wsse="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd" SOAP-ENV:mustUnderstand="1"><ds:Signature xmlns:ds="http://www.w3.org/2000/09/xmldsig#"><ds:SignedInfo><ds:CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"></ds:CanonicalizationMethod><ds:SignatureMethod Algorithm="http://www.w3.org/2000/09/xmldsig#rsa-sha1"></ds:SignatureMethod><ds:Reference URI="#%(body_id)s"><ds:Transforms><ds:Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"></ds:Transform></ds:Transforms><ds:DigestMethod Algorithm="http://www.w3.org/2000/09/xmldsig#sha1"></ds:DigestMethod><ds:DigestValue>%(digest_value)s</ds:DigestValue></ds:Reference></ds:SignedInfo><ds:SignatureValue>%(signature_value)s</ds:SignatureValue><ds:KeyInfo><wsse:SecurityTokenReference><ds:X509Data><ds:X509IssuerSerial><ds:X509IssuerName>%(issuer_name)s</ds:X509IssuerName><ds:X509SerialNumber>%(serial_number)s</ds:X509SerialNumber></ds:X509IssuerSerial></ds:X509Data></wsse:SecurityTokenReference></ds:KeyInfo></ds:Signature></wsse:Security></SOAP-ENV:Header><SOAP-ENV:Body xmlns:wsu="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd" wsu:Id="%(body_id)s"><ns1:%(action)s xmlns:ns1="http://webservices.webpayserver.transbank.com/"><arg0>%(params)s</arg0></ns1:%(action)s></SOAP-ENV:Body></SOAP-ENV:Envelope>"""

BODY_TMPL = """<SOAP-ENV:Body xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/" xmlns:wsu="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd" wsu:Id="%(body_id)s"><ns1:%(action)s xmlns:ns1="http://webservices.webpayserver.transbank.com/"><arg0>%(params)s</arg0></ns1:%(action)s></SOAP-ENV:Body>"""


class Document(object):
    _key = None
    _cert = None
    _x509 = None
    _action = None
    _params = None
    doc = None

    def __init__(self, action, params):
        self._action = action
        self._params = params
        self.doc = self.build_doc()

    @property
    def key(self):
        if not self._key:
            self._key = os.getenv('TBK_COMMERCE_KEY')
        return self._key

    @property
    def cert(self):
        if not self._cert:
            self._cert = os.getenv('TBK_COMMERCE_CRT')
        return self._cert

    @property
    def x509(self):
        if not self._x509:
            self._x509 = crypto.load_certificate(crypto.FILETYPE_PEM, 
                                                 open(self.cert).read())
        return self._x509

    def get_issuer_name(self):
        issuer = self.x509.get_issuer()
        component_list = []
        for component in issuer.get_components():
            component_list.append('{}={}'.format(component[0], component[1]))
        return ", ".join(component_list)

    def get_serial_number(self):
        return str(self.x509.get_serial_number())

    def get_digest_value(self, xml, c14n_exc=True):
        return base64.b64encode(rsa.pkcs1._hash(xml, 'SHA-1'))

    def get_body_id(self):
        m = md5.new()
        m.update("{}{}{}".format(self._action, self._params,
                 arrow.now().format('YYYY-MM-DD HH:mm:ss ZZ')))
        return m.hexdigest()

    def rsa_sign(self, xml):
        with open(self.key) as privatefile:
            keydata = privatefile.read()
            key = rsa.PrivateKey.load_pkcs1(keydata)
            signature = rsa.sign(xml, key, 'SHA-1')
            return base64.b64encode(signature)

    def build_params_xml(self, params):
        params_xml = ""
        for k, v in params.items():
            params_xml += '<{0}>{1}</{0}>'.format(k, v)
        return params_xml

    def build_doc(self):
        body_id = self.get_body_id()
        # 1) build body
        body_params = self.build_params_xml(self._params)
        params = {'action': self._action, 'params': body_params, 'body_id': body_id}
        body = BODY_TMPL % params
        # 2) firm with body
        digest_value = self.get_digest_value(body, True)
        # 3) assign
        xml_to_sign = SIGN_ENV_TMPL % {'digest_value': digest_value, 'body_id': body_id}
        signature_value = self.rsa_sign(xml=xml_to_sign)
        # get params
        params = {'signature_value': signature_value, 'issuer_name': self.get_issuer_name(),
                  'serial_number': self.get_serial_number(), 'digest_value': digest_value,
                  'body_id': body_id, 'action': self._action, 'params': body_params}

        # 4) build headers
        return XML_TMPL % params



#!/usr/bin/env python

from __future__ import print_function
import struct
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.poolmanager import PoolManager
import ssl

try:
    range = xrange
except NameError:
    pass


def _left_rotate(n, b):
    return ((n << b) | (n >> (32 - b))) & 0xffffffff


def sha1(message):
    """SHA-1 Hashing Function

    A custom SHA-1 hashing function implemented entirely in Python.

    Arguments:
        message: The input message string to hash.

    Returns:
        A hex SHA-1 digest of the input message.
    """
    # Initialize variables:
    h0 = 0x67452301
    h1 = 0xEFCDAB89
    h2 = 0x98BADCFE
    h3 = 0x10325476
    h4 = 0xC3D2E1F0
    
    # Pre-processing:
    original_byte_len = len(message)
    original_bit_len = original_byte_len * 8
    # append the bit '1' to the message
    message += b'\x80'
    
    # append 0 <= k < 512 bits '0', so that the resulting message length (in bits)
    #    is congruent to 448 (mod 512)
    message += b'\x00' * ((56 - (original_byte_len + 1) % 64) % 64)
    
    # append length of message (before pre-processing), in bits, as 64-bit big-endian integer
    message += struct.pack(b'>Q', original_bit_len)
    # Process the message in successive 512-bit chunks:
    # break message into 512-bit chunks
    for i in range(0, len(message), 64):
        w = [0] * 80
        # break chunk into sixteen 32-bit big-endian words w[i]
        for j in range(16):
            w[j] = struct.unpack(b'>I', message[i + j*4:i + j*4 + 4])[0]
        # Extend the sixteen 32-bit words into eighty 32-bit words:
        for j in range(16, 80):
            w[j] = _left_rotate(w[j-3] ^ w[j-8] ^ w[j-14] ^ w[j-16], 1)
    
        # Initialize hash value for this chunk:
        a = h0
        b = h1
        c = h2
        d = h3
        e = h4
    
        for i in range(80):
            if 0 <= i <= 19:
                # Use alternative 1 for f from FIPS PB 180-1 to avoid ~
                f = d ^ (b & (c ^ d))
                k = 0x5A827999
            elif 20 <= i <= 39:
                f = b ^ c ^ d
                k = 0x6ED9EBA1
            elif 40 <= i <= 59:
                f = (b & c) | (b & d) | (c & d) 
                k = 0x8F1BBCDC
            elif 60 <= i <= 79:
                f = b ^ c ^ d
                k = 0xCA62C1D6
    
            a, b, c, d, e = ((_left_rotate(a, 5) + f + e + k + w[i]) & 0xffffffff, 
                            a, _left_rotate(b, 30), c, d)
    
        # sAdd this chunk's hash to result so far:
        h0 = (h0 + a) & 0xffffffff
        h1 = (h1 + b) & 0xffffffff 
        h2 = (h2 + c) & 0xffffffff
        h3 = (h3 + d) & 0xffffffff
        h4 = (h4 + e) & 0xffffffff
    
    # Produce the final hash value (big-endian):
    return '%08x%08x%08x%08x%08x' % (h0, h1, h2, h3, h4)


from response import Response
import requests
import sys

try:  # patch urllib with pyopenssl for TLS 1.2 compatibility (python 2.6)
    import urllib3.contrib.pyopenssl
    urllib3.contrib.pyopenssl.inject_into_urllib3()
except ImportError:
    pass

if sys.version > '3':
    unicode = str


class Client(object):
    client = None
    location = None
    _testing_mode = None
    http = None

    def __init__(self, testing_mode=False):
        self._testing_mode = testing_mode
        if testing_mode:
            self.location = 'https://tbk.orangepeople.cl/webpayserver/wswebpay/OneClickPaymentService'
        else:
            self.location = 'https://webpay3g.transbank.cl:443/webpayserver/wswebpay/OneClickPaymentService'

    def send(self, action, xml):
        soap_action = str(action)

        headers = {
            'Content-type': 'text/xml; charset="UTF-8"',
            'Content-length': str(len(xml)),
            'SOAPAction': '"%s"' % soap_action
        }

        if sys.version < '3':
            headers = dict((str(k), str(v)) for k, v in headers.items())

        d = requests.post(self.location, data=xml, headers=headers)
        return d.text

    def request(self, action, xml):
        response_content = self.send(action, xml)
        response = Response(response_content, action, self._testing_mode)
        return response



# encoding=UTF-8
from __future__ import unicode_literals

import os
import datetime
from contextlib import closing
import pytz


LOGGLY_LOG_NAME = "OneClick"
EVENTS_LOG_FILE_NAME_FORMAT = "TBK_EVN%s.log"
EVENTS_LOG_FILE_DATE_FORMAT = "%Y%m%d"


class SimpleHandler(object):
    def __init__(self, path=None):
        self.path = path

    def event_generic(self, **kwargs):
        event_type = kwargs.pop('type')
        event_info = '{} ({}) \n'.format(event_type, kwargs['action'])

        for k, v in kwargs.items():
            event_info = '{}\t{} => {}\n'.format(event_info, k, v)

        with closing(self.events_log_file) as events_log_file:
            events_log_file.write(event_info)

    @property
    def events_log_file(self):
        return self.log_file(EVENTS_LOG_FILE_NAME_FORMAT, EVENTS_LOG_FILE_DATE_FORMAT)

    def log_file(self, log_file_name_format, log_file_date_format):
        santiago = pytz.timezone('America/Santiago')
        now = santiago.localize(datetime.datetime.now())
        file_name = log_file_name_format % now.strftime(log_file_date_format)
        return open(os.path.join(self.path, file_name), 'a+')


# -*- coding: utf-8 -*-
import xml.etree.ElementTree as ET
from OpenSSL import crypto
from cStringIO import StringIO
import os

RESPONSE_CODE = {
    'Authorize': {
        '0': 'aprobado',
        '-1': 'rechazo de transacción',
        '-2': 'rechazo, intenténlo nuevamente',
        '-3': 'rechazo, error en transacción',
        '-4': 'rechazo de transacción',
        '-5': 'rechazo por error de tasa',
        '-6': 'rechazo, excede monto máximo mensual',
        '-7': 'rechazo, excede límite diario por transacción',
        '-8': 'rechazo, rubro no autorizado',
        '-97': 'limites Oneclick, máximo monto diario de pago excedido',
        '-98': 'limites Oneclick, máximo monto de pago excedido',
        '-99': 'limites Oneclick, máxima cantidad de pagos diarios excedido'
    },
    'default': {
        '0': 'aprobado',
        '-98': 'Error'
    }
}

VALID_RESPONSE_PARAMS = {
    'Authorize': ['authorizationCode', 'creditCardType',
                  'last4CardDigits', 'responseCode', 'transactionId'],
    'initInscription': ['token', 'urlWebpay'],
    'finishInscription': ['authCode', 'creditCardType', 'last4CardDigits',
                          'responseCode', 'tbkUser'],
    'codeReverseOneClick': ['reverseCode', 'reversed'],
    'removeUser': ['removed']
}


class Response(object):
    action = None
    content = None
    xml_response = None
    error = None
    error_msg = ''
    user_error_msg = ''
    extra = {}
    _xml_result = None
    _xml_error = None
    _tbk_key = None
    _testing_mode = None

    def __init__(self, content, action, testing_mode=False):
        self.error = None
        self._testing_mode = testing_mode
        self.content = self._canonicalize(content)
        self.action = action
        self.xml_response = self.build_xml_response(content)
        self.validate()

    def build_xml_response(self, xml_string):
        try:
            return ET.fromstring(xml_string)
        except ET.ParseError:
            return None

    def _canonicalize(self, xml):  # TODO: move to utils or document.p
        output = StringIO()
        from .simplexml import SimpleXMLElement
        try:
            SimpleXMLElement(xml).write_c14n(output, exclusive=True)
            return output.getvalue()
        except:  # TODO: handle exceptions
            return None

    @property
    def tbk_key(self):
        if not self._tbk_key:
            self._tbk_key = crypto.load_certificate(crypto.FILETYPE_PEM,
                                                    open(os.getenv('TBK_PUBLIC_CRT')).read())
        return self._tbk_key

    @property
    def _signed_info(self):  # TODO: move to utils or document.py
        namespaces = ['{http://schemas.xmlsoap.org/soap/envelope/}Header',
                      '{http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd}Security',
                      '{http://www.w3.org/2000/09/xmldsig#}Signature',
                      '{http://www.w3.org/2000/09/xmldsig#}SignedInfo']
        element = self.xml_response
        for ns in namespaces:
            element = element.findall(ns)[0]
        signed_info = ET.tostring(element)
        return self._canonicalize(signed_info)

    @property
    def _signature_value(self):
        namespaces = ['{http://schemas.xmlsoap.org/soap/envelope/}Header',
                      '{http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd}Security',
                      '{http://www.w3.org/2000/09/xmldsig#}Signature',
                      '{http://www.w3.org/2000/09/xmldsig#}SignatureValue']

        element = self.xml_response
        for ns in namespaces:
            element = element.findall(ns)[0]
        signature_value = element.text
        return signature_value

    def _is_valid_signature(self):
        if self._testing_mode:  # skip validation on testing mode
            return True
        elif not os.getenv('TBK_PUBLIC_CRT'):  # tbk certificate undefined
            return True

        try:  # OpenSSL return: None if the signature is correct, raise exception otherwise
            crypto.verify(self.tbk_key, self._signature_value, self._signed_info, 'sha1')
            return True
        except:
            return False

    def str2bool(self, bool_string):  # TODO: move to utils
        if bool_string.lower() == 'true':
            return True
        elif bool_string.lower() == 'false':
            return False
        raise TypeError

    @property
    def params(self):
        if not self._xml_result:
            result = {}
            for e in self.xml_response.findall('.//return'):
                for children in e.getchildren():
                    if self.action == 'codeReverseOneClick':
                        try:
                            result[children.tag] = self.str2bool(children.text)
                        except TypeError:
                            result[children.tag] = children.text
                    else:
                        result[children.tag] = children.text
                else:
                    if self.action == 'removeUser':
                        result['removed'] = self.str2bool(e.text)

            obj = {}
            for p in VALID_RESPONSE_PARAMS[self.action]:
                obj[p] = result.get(p)
                setattr(self, p, result.get(p))
            self._xml_result = obj
        return self._xml_result

    @property
    def xml_error(self):
        self._xml_error = None
        if not self._xml_error and self.xml_response:
            faultcode = self.xml_response.findall('.//faultcode')
            faultstring = self.xml_response.findall('.//faultstring')
            if faultcode and faultstring:
                self._xml_error = {'faultcode': faultcode[0].text,
                                   'faultstring': faultstring[0].text}
        return self._xml_error

    def is_valid(self):  # True or False if any errors occurred (exceptions included)
        if self.error:
            return False
        return True

    @property
    def response_code(self):
        if self.params and 'responseCode' in self.params:
            return str(self.params['responseCode'])
        return None

    def response_code_display(self):
        if self.action in RESPONSE_CODE and self.response_code in RESPONSE_CODE[self.action]:
            return RESPONSE_CODE[self.action][self.response_code]
        elif self.response_code in RESPONSE_CODE['default']:
            return RESPONSE_CODE['default'][self.response_code]
        else:
            return self.response_code

    def validate(self):
        if not self.xml_response:
            self.error = 'SoapServerError'
            self.error_msg = 'invalid XML response'
            self.user_error_msg = 'Error al realizar la operación'
            self.extra = {}
        elif self.xml_error:
            self.error = 'SoapServerError'
            self.error_msg = self.xml_error['faultstring']
            self.user_error_msg = 'Error al realizar la operación'
            self.extra = self.xml_error
        elif not self._is_valid_signature():
            self.error = 'SecurityError'
            self.error_msg = 'invalid signature value'
            self.user_error_msg = 'Error al realizar la operación'
            self.extra = self.xml_error
        else:
            if self.action in ['finishInscription', 'Authorize'] and int(self.response_code) != 0:
                self.error = '{}Error'.format(self.action)
                self.error_msg = self.response_code_display()
                self.user_error_msg = self.error_msg
                self.extra = {'response_code': self.response_code}
            elif self.action == 'removeUser' and not self.params['removed']:
                self.error = 'removeUserError'
                self.error_msg = 'imposible eliminar la inscripción'
                self.user_error_msg = self.error_msg
                self.extra = {'removed': False}
            elif self.action == 'codeReverseOneClick' and not self.params['reversed']:
                self.error = 'codeReverseOneClickError'
                self.error_msg = 'imposible revertir la compra'
                self.user_error_msg = self.error_msg
                self.extra = {'reversed': False}



#!/usr/bin/python
# -*- coding: utf-8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""Simple XML manipulation"""

from __future__ import unicode_literals
import sys
if sys.version > '3':
    basestring = str
    unicode = str

import re
import xml.dom.minidom


class SimpleXMLElement(object):
    """Simple XML manipulation (simil PHP)"""

    def __init__(self, text=None, elements=None, document=None,
                 namespace=None, prefix=None, namespaces_map={}, jetty=False):
        """
        :param namespaces_map: How to map our namespace prefix to that given by the client;
          {prefix: received_prefix}
        """
        self.__namespaces_map = namespaces_map
        _rx = "|".join(namespaces_map.keys())  # {'external': 'ext', 'model': 'mod'} -> 'external|model'
        self.__ns_rx = re.compile(r"^(%s):.*$" % _rx)  # And now we build an expression ^(external|model):.*$
                                                       # to find prefixes in all xml nodes i.e.: <model:code>1</model:code>
                                                       # and later change that to <mod:code>1</mod:code>
        self.__ns = namespace
        self.__prefix = prefix
        self.__jetty = jetty                           # special list support

        if text is not None:
            try:
                self.__document = xml.dom.minidom.parseString(text)
            except:
                raise
            self.__elements = [self.__document.documentElement]
        else:
            self.__elements = elements
            self.__document = document

    def add_child(self, name, text=None, ns=True):
        """Adding a child tag to a node"""
        if not ns or self.__ns is False:
            element = self.__document.createElement(name)
        else:
            if isinstance(ns, basestring):
                element = self.__document.createElement(name)
                if ns:
                    element.setAttribute("xmlns", ns)
            elif self.__prefix:
                element = self.__document.createElementNS(self.__ns, "%s:%s" % (self.__prefix, name))
            else:
                element = self.__document.createElementNS(self.__ns, name)
        # don't append null tags!
        if text is not None:
            if isinstance(text, xml.dom.minidom.CDATASection):
                element.appendChild(self.__document.createCDATASection(text.data))
            else:
                element.appendChild(self.__document.createTextNode(text))
        self._element.appendChild(element)
        return SimpleXMLElement(
            elements=[element],
            document=self.__document,
            namespace=self.__ns,
            prefix=self.__prefix,
            jetty=self.__jetty,
            namespaces_map=self.__namespaces_map
        )

    def __setattr__(self, tag, text):
        """Add text child tag node (short form)"""
        if tag.startswith("_"):
            object.__setattr__(self, tag, text)
        else:
            self.add_child(tag, text)

    def __delattr__(self, tag):
        """Remove a child tag (non recursive!)"""
        elements = [__element for __element in self._element.childNodes
                    if __element.nodeType == __element.ELEMENT_NODE]
        for element in elements:
            self._element.removeChild(element)

    def add_comment(self, data):
        """Add an xml comment to this child"""
        comment = self.__document.createComment(data)
        self._element.appendChild(comment)

    def as_xml(self, filename=None, pretty=False):
        """Return the XML representation of the document"""
        if not pretty:
            return self.__document.toxml('UTF-8')
        else:
            return self.__document.toprettyxml(encoding='UTF-8')

    if sys.version > '3':
        def __repr__(self):
            """Return the XML representation of this tag"""
            return self._element.toxml()
    else:
        def __repr__(self):
            """Return the XML representation of this tag"""
            # NOTE: do not use self.as_xml('UTF-8') as it returns the whole xml doc
            return self._element.toxml('UTF-8')

    def get_name(self):
        """Return the tag name of this node"""
        return self._element.tagName

    def get_local_name(self):
        """Return the tag local name (prefix:name) of this node"""
        return self._element.localName

    def get_prefix(self):
        """Return the namespace prefix of this node"""
        return self._element.prefix

    def get_namespace_uri(self, ns):
        """Return the namespace uri for a prefix"""
        element = self._element
        while element is not None and element.attributes is not None:
            try:
                return element.attributes['xmlns:%s' % ns].value
            except KeyError:
                element = element.parentNode

    def attributes(self):
        """Return a dict of attributes for this tag"""
        #TODO: use slice syntax [:]?
        return self._element.attributes

    def __getitem__(self, item):
        """Return xml tag attribute value or a slice of attributes (iter)"""
        if isinstance(item, basestring):
            if self._element.hasAttribute(item):
                return self._element.attributes[item].value
        elif isinstance(item, slice):
            # return a list with name:values
            return list(self._element.attributes.items())[item]
        else:
            # return element by index (position)
            element = self.__elements[item]
            return SimpleXMLElement(
                elements=[element],
                document=self.__document,
                namespace=self.__ns,
                prefix=self.__prefix,
                jetty=self.__jetty,
                namespaces_map=self.__namespaces_map
            )

    def add_attribute(self, name, value):
        """Set an attribute value from a string"""
        self._element.setAttribute(name, value)

    def __setitem__(self, item, value):
        """Set an attribute value"""
        if isinstance(item, basestring):
            self.add_attribute(item, value)
        elif isinstance(item, slice):
            # set multiple attributes at once
            for k, v in value.items():
                self.add_attribute(k, v)

    def __delitem__(self, item):
        "Remove an attribute"
        self._element.removeAttribute(item)

    def __call__(self, tag=None, ns=None, children=False, root=False,
                 error=True, ):
        """Search (even in child nodes) and return a child tag by name"""
        try:
            if root:
                # return entire document
                return SimpleXMLElement(
                    elements=[self.__document.documentElement],
                    document=self.__document,
                    namespace=self.__ns,
                    prefix=self.__prefix,
                    jetty=self.__jetty,
                    namespaces_map=self.__namespaces_map
                )
            if tag is None:
                # if no name given, iterate over siblings (same level)
                return self.__iter__()
            if children:
                # future: filter children? by ns?
                return self.children()
            elements = None
            if isinstance(tag, int):
                # return tag by index
                elements = [self.__elements[tag]]
            if ns and not elements:
                for ns_uri in isinstance(ns, (tuple, list)) and ns or (ns, ):
                    elements = self._element.getElementsByTagNameNS(ns_uri, tag)
                    if elements:
                        break
            if self.__ns and not elements:
                elements = self._element.getElementsByTagNameNS(self.__ns, tag)
            if not elements:
                elements = self._element.getElementsByTagName(tag)
            if not elements:
                if error:
                    raise AttributeError("No elements found")
                else:
                    return
            return SimpleXMLElement(
                elements=elements,
                document=self.__document,
                namespace=self.__ns,
                prefix=self.__prefix,
                jetty=self.__jetty,
                namespaces_map=self.__namespaces_map)
        except AttributeError as e:
            raise AttributeError("Tag not found: %s (%s)" % (tag, e))

    def __getattr__(self, tag):
        """Shortcut for __call__"""
        return self.__call__(tag)

    def __iter__(self):
        """Iterate over xml tags at this level"""
        try:
            for __element in self.__elements:
                yield SimpleXMLElement(
                    elements=[__element],
                    document=self.__document,
                    namespace=self.__ns,
                    prefix=self.__prefix,
                    jetty=self.__jetty,
                    namespaces_map=self.__namespaces_map)
        except:
            raise

    def __dir__(self):
        """List xml children tags names"""
        return [node.tagName for node
                in self._element.childNodes
                if node.nodeType != node.TEXT_NODE]

    def children(self):
        """Return xml children tags element"""
        elements = [__element for __element in self._element.childNodes
                    if __element.nodeType == __element.ELEMENT_NODE]
        if not elements:
            return None
            #raise IndexError("Tag %s has no children" % self._element.tagName)
        return SimpleXMLElement(
            elements=elements,
            document=self.__document,
            namespace=self.__ns,
            prefix=self.__prefix,
            jetty=self.__jetty,
            namespaces_map=self.__namespaces_map
        )

    def __len__(self):
        """Return element count"""
        return len(self.__elements)

    def __contains__(self, item):
        """Search for a tag name in this element or child nodes"""
        return self._element.getElementsByTagName(item)

    def __unicode__(self):
        """Returns the unicode text nodes of the current element"""
        rc = ''
        for node in self._element.childNodes:
            if node.nodeType == node.TEXT_NODE or node.nodeType == node.CDATA_SECTION_NODE:
                rc = rc + node.data
        return rc

    if sys.version > '3':
        __str__ = __unicode__
    else:
        def __str__(self):
            return self.__unicode__().encode('utf-8')

    def __int__(self):
        """Returns the integer value of the current element"""
        return int(self.__str__())

    def __float__(self):
        """Returns the float value of the current element"""
        try:
            return float(self.__str__())
        except:
            raise IndexError(self._element.toxml())

    _element = property(lambda self: self.__elements[0])

    def _update_ns(self, name):
        """Replace the defined namespace alias with tohse used by the client."""
        pref = self.__ns_rx.search(name)
        if pref:
            pref = pref.groups()[0]
            try:
                name = name.replace(pref, self.__namespaces_map[pref])
            except KeyError:
                pass
        return name

    def import_node(self, other):
        x = self.__document.importNode(other._element, True)  # deep copy
        self._element.appendChild(x)

    def write_c14n(self, output=None, exclusive=True):
        "Generate the canonical version of the XML node"
        from . import c14n
        xml = c14n.Canonicalize(self._element, output,
                                unsuppressedPrefixes=[] if exclusive else None)
        return xml



__all__ = ['logger', 'BaseHandler', 'NullHandler']


LOG_DATE_FORMAT = "%d%m%Y"
LOG_TIME_FORMAT = "%H%M%S"


class BaseHandler(object):

    def event_generic(self, **kwargs):
        raise NotImplementedError()


class NullHandler(BaseHandler):

    def event_generic(self, **kwargs):
        pass


class Logger(object):

    def __init__(self, handler):
        self.set_handler(handler)

    def set_handler(self, handler):
        self.handler = handler

    def generic(self, action, request, response):
        # request
        params = request.params
        params['action'] = action
        params['type'] = 'REQUEST'
        self.handler.event_generic(**params)
        # response, TODO: handle error
        if response.is_valid():
            params = response.params
            params['type'] = 'RESPONSE'
            params['action'] = action
            self.handler.event_generic(**params)
        else:
            params = {'error': response.error, 'error_msg': response.error_msg,
                      'user_error_msg': response.user_error_msg, 'extra': response.extra}
            params['type'] = 'RESPONSE'
            params['action'] = action
            self.handler.event_generic(**params)


logger = Logger(NullHandler())


def configure_logger(handler):
    logger.set_handler(handler)



# encoding=UTF-8
#!/usr/bin/env python
import os
from setuptools import setup, find_packages

setup(
    name='tbk-oneclick',
    version='0.8',
    description='Python implementation of transbank oneclick',
    long_description='Python implementation of transbank oneclick',
    author='Ignacio Hermosilla',
    author_email='hi@ignacio.im',
    maintainer='Ignacio  Hermosilla',
    maintainer_email='hi@ignacio.im',
    licence='MIT',
    url='https://github.com/ignacioHermosilla/tbk-oneclick',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
    ],
    keywords='transbank oneclick one_click webpay tbk chile payments pago',
    install_requires=[
        'pyopenssl',
        'arrow',
        'rsa',
        'requests',
        'urllib3',
        'pyasn1',
        'ndg-httpsclient'],
    tests_require=[
        'mock>=1.0.1',
        'nose>=1.3.3',
    ],
    packages=find_packages(),
    test_suite='nose.collector',
    zip_safe=True,
)


from document import Document
from request import Request
from client import Client
from .logging import logger


class OneClick(object):
    client = None

    def __init__(self, testing_mode=False):
        self.client = Client(testing_mode)

    def init_inscription(self, email, response_url, username):
        params = {'email': email, 'username': username,
                  'responseURL': response_url}
        request = Request(**params)
        d = Document(action='initInscription', params=params)
        response = self.client.request('initInscription', d.doc)
        logger.generic('init_inscription', request, response)
        return response

    def finish_inscription(self, token):
        params = {'token': token}
        request = Request(**params)
        d = Document(action='finishInscription', params=params)
        response = self.client.request('finishInscription', d.doc)
        logger.generic('finish_inscription', request, response)
        return response

    def authorize(self, amount, tbk_user, username, buy_order):
        params = {'amount': amount, 'tbkUser': tbk_user,
                  'username': username, 'buyOrder': buy_order}
        request = Request(**params)
        d = Document(action='authorize', params=params)
        response = self.client.request('Authorize', d.doc)
        logger.generic('authorize', request, response)
        return response

    def reverse(self, buy_order):
        params = {'buyorder': buy_order}
        request = Request(**params)
        d = Document(action='codeReverseOneClick', params=params)
        response = self.client.request('codeReverseOneClick', d.doc)
        logger.generic('reverse', request, response)
        return response

    def remove_user(self, tbk_user, username):
        params = {'tbkUser': tbk_user, 'username': username}
        request = Request(**params)
        d = Document(action='removeUser', params=params)
        response = self.client.request('removeUser', d.doc)
        logger.generic('remove_user', request, response)
        return response
